:- module(analyser, [analyser/0]).
:- use_module(library(lists)).
:- use_module(library(timeout)).
:- use_module([abstract, benchmarks, calibrate, fixpoint, pretty, pure]).

analyser :-
	format("program  [air, ..., zebra] : ", []),
	read(Prog),
	findall(Row, analyser(Prog, _, _, Row), Table),
	ascii_table(Table),
	tell('../results/results.tex'),
	ascii_table(Table),
	told,
	true.

analyser(Prog, _, Domain, [Prog, Type, Domain | Stats]) :-
	benchmark(Prog, Type, Goal, Share, Pos_Goal), 
	domain_extension(Domain, _),
	analyser_aux(Prog, Type, Domain, Goal, Share, Pos_Goal, Stats).

ascii_table(Table) :-
	format("\\documentstyle[a4wide]{article}", []),
	format("~n\\begin{document}", []),
	format("~n\\def\\widdly{\\hspace{4pt}}", []),
	format("~n\\scriptsize", []),
	format("~n\\begin{tabular}{r|r|r|", []),
	format("~nr@{\\widdly}r@{\\widdly}r@{\\widdly}r|", []),
	format("~nr@{\\widdly}r@{\\widdly}r@{\\widdly}r|", []),
	format("~nr@{\\widdly}r@{\\widdly}r}", []),
	format("~n& & & \\multicolumn{4}{c}{fixpoint}", []),
	format("~n& \\multicolumn{4}{|c}{precision}", []),
	format("~n& \\multicolumn{3}{|c}{\\% precision} \\\\", []),
	format("~nfile & size & abs ", []),
	format("~n& con & det & msg & pos ", []),
	format("~n& con & det & msg & pos ", []),
	format("~n& con & det & msg ", []),
	ascii_table(Table, dummy_file_name, [], []),
	format("~n\\end{tabular}", []),
	format("~n\\end{document}", []).

ascii_table([], _, Acc_Fix, Acc_Count) :-
	rev_time_list(Acc_Fix, '$\\infty$'),
	format("~n", []),
	rev_precision_list(Acc_Count, '--'),
	format("~n", []),
	relative_precision_list(Acc_Count, Percent_Count),
	rev_precision_list(Percent_Count, '--'),
	format("\\\\ ", []).
ascii_table([Row | Table], File, Acc_Fix, Acc_Count) :-
	Row = [File, _, _, _, _, Fix, _, Count],
	!,
	ascii_table(Table, File, [Fix | Acc_Fix], [Count | Acc_Count]).
ascii_table([Row | Table], _, Acc_Fix, Acc_Count) :-
	rev_time_list(Acc_Fix, '$\\infty$'),
	format("~n", []),
	rev_precision_list(Acc_Count, '--'),
	format("~n", []),
	relative_precision_list(Acc_Count, Percent_Count),
	rev_precision_list(Percent_Count, '--'),
	format("\\\\ ~n", []),
	Row = [New_File, New_Type, _, New_Size, New_Abs, New_Fix, _, New_Count],
	name(New_File, Ascii_File),
	ascii_file(Ascii_File),
	(New_Type = herbrand -> format(".pl", []); format(".clpr", [])),
	format("~n& ~w & ~2d~n", [New_Size, New_Abs]),
	ascii_table(Table, New_File, [New_Fix], [New_Count]).

ascii_file([]).
ascii_file([95 | L]) :-
	!,
	format("\\_", []),
	ascii_file(L).
ascii_file([C | L]) :-
	format("~c", [C]),
	ascii_file(L).

rev_time_list([], _).
rev_time_list([inf | Ns], Symbol) :-
	!,
	rev_time_list(Ns, Symbol),
	format("& ~w ", [Symbol]).
rev_time_list([N | Ns], Symbol) :-
	rev_time_list(Ns, Symbol),
	format("& ~2d ", [N]).

rev_precision_list([], _).
rev_precision_list([inf | Ns], Symbol) :-
	!,
	rev_precision_list(Ns, Symbol),
	format("& ~w ", [Symbol]).
rev_precision_list([N | Ns], Symbol) :-
	rev_precision_list(Ns, Symbol),
	format("& ~w ", [N]).

relative_precision_list([], []).
relative_precision_list([Best | Counts], Percents) :-
	relative_precision_list(Counts, Best, Percents).

relative_precision_list([], _, []).
relative_precision_list([inf | Counts], Best, [inf | Percents]) :-
	!,
	relative_precision_list(Counts, Best, Percents).
relative_precision_list([Count | Counts], Best, [Percent | Percents]) :-
	(number(Best), Best > 0 ->
		Percent is integer(round((100 * (Best - Count)) / Best));
		Percent = 0),
	relative_precision_list(Counts, Best, Percents).

analyser_aux(Prog, Type, Domain, Goal, Share, Pos_Goal, [Size_Abs, Time_Abs, Time_Fix, Time_Total, Count]) :-
	name(Prog, SProg),
	append("../benchmarks/", SProg, SBProg),
	append("../results/", SProg, SRProg),
	type_extension(Type, SType),
	append(SBProg, SType, SIFile),
	name(IFile, SIFile),
	domain_extension(Domain, SDomain),
	append(SRProg, SDomain, SOFile),
	name(OFile, SOFile),
	garbage_collect,
	open(IFile, read, IStream),
	format("{abstracting ~w... ", [IFile]),
	ttyflush,
	statistics(runtime, _),
    	abstract(IStream, Type, Size_Abs), 
	close(IStream),
        statistics(runtime, [_, Raw_Time_Abs]), 
	Time_Abs is Raw_Time_Abs // 10,
	format("~2d sec}~n", [Time_Abs]),
	append(SRProg, ".pl", PSFile),
	name(PFile, PSFile),
	open(PFile, write, OStream),
	format("{purifying ~w...", [IFile]),
	ttyflush,
    	purify(Prog, OStream), 
	format("}~n", []),
	close(OStream),
	statistics(runtime, _),
	format("{iterating with ~w... ", [Domain]), 
	ttyflush,
	Minutes is 15,
	Milli is integer(Minutes * 60 * 1000),	
	time_out(fixpoint:fixpoint(td, Domain, Goal, Share, Pos_Goal), Milli, Success),
	analyser_aux_aux(Success, OFile, Domain, [Time_Abs, Time_Fix, Time_Total, Count]).

analyser_aux_aux(success, OFile, Domain, [Time_Abs, Time_Fix, Time_Total, Count]) :-
	statistics(runtime, [_, Raw_Time_Fix]),
	Time_Fix is Raw_Time_Fix // 10,
	format("~2d sec}~n", [Time_Fix]),
	Time_Total is Time_Abs + Time_Fix,
	open(OFile, write, OStream),
	time_out(pretty:pretty_print(Domain, OStream), 10000, Success),
%	pretty_print(Domain, OStream), Success = success,
	close(OStream),
	analyser_aux_aux_aux(Success, OFile, Domain, Count).
analyser_aux_aux(time_out, _, _, [_, inf, inf, inf]) :-
	statistics(runtime, [_, Raw_Time_Fix]),
	Time_Fix is Raw_Time_Fix // 10,
	format("> ~2d sec}~n", [Time_Fix]).

analyser_aux_aux_aux(success, OFile, Domain, Count) :-
	calibrate(Domain, Count),
	format("{calibrating ~w... ~w ground arguments}~n", [OFile, Count]).
analyser_aux_aux_aux(time_out, OFile, Domain, Count) :-
	open(OFile, write, OStream),
	close(OStream),
	analyser_aux_aux_aux(success, OFile, Domain, Count).

type_extension(herbrand, ".pl").
type_extension(numeric, ".clpr").

domain_extension(con_nonground_model, ".con_nonground_model").
domain_extension(det_nonground_model, ".det_nonground_model").
domain_extension(msg_nonground_model, ".msg_nonground_model").
domain_extension(pos_nonground_model, ".pos_nonground_model").
